__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

import sys
import os
import select

class CECPProtocol:
    """CECP protocol en/decoder.
    
    
    """
    
    # Data being accumulated to be parsed
    __buffer = ''
    
    NEWLINE             = '\n'
    MOVE_PREFIXS        = ['My move is: ', 'my move is ', 'move ']
    INVALID_MOVE_PREFIX = 'Illegal move: '
    RESIGN_PREFIX       = 'tellics resign'
    DRAW_PREFIX         = '1/2-1/2'
    
    def __init__(self):
        """
        """
        # Go to simple interface mode
        self.onOutgoingData('xboard\n')
        
        # TEMP: Play nice
        self.sendSetPondering(False)
        #self.sendSetSearchDepth(4)
        
    # Methods to extend
    
    def onOutgoingData(self, data):
        """Called when there is data to send to the CECP engine.
        
        'data' is the data to give to the AI (string).
        """
        print 'OUT: ' + repr(data)
    
    def onUnknownLine(self, line):
        """Called when an unknown line is received from the CECP AI.
        
        'line' is the line that has not been decoded (string). There is
               no newline on the end of the string.
        """
        print 'Unknown CECP line: ' + line
    
    def onMove(self, move):
        """Called when the AI makes a move.
        
        'move' is the move the AI has decided to make (string).
        """
        print 'CECP move: ' + move
    
    def onIllegalMove(self, move):
        """Called when the AI rejects a move.
        
        'move' is the move the AI rejected (string).
        """
        print 'CECP illegal move: ' + move
        
    def onResign(self):
        """Called when the AI resigns"""
        print 'CECP AI resigns'
        
    # Public methods
    
    def sendSetSearchDepth(self, searchDepth):
        """Set the search depth for the AI.
        
        'searchDepth' is the number of moves to look ahead (integer).
        """
        # This is the CECP specified method
        self.onOutgoingData('sd %i\n' % int(searchDepth))
        
        # GNUchess uses this instead
        self.onOutgoingData('depth %i\n' % int(searchDepth))
        
    def sendSetPondering(self, aiPonders):
        """Enable/disable AI pondering.
        
        'aiPonders' is a flag to show if the AI thinks during opponent moves (True) or not (False).
        """
        if aiPonders:
            self.onOutgoingData('hard\n')
        else:
            self.onOutgoingData('easy\n')
    
    def sendMove(self, move):
        """Move for the current player.
        
        'move' is the move the current player has made (string).
        """
        self.onOutgoingData(move + '\n')
        
    def sendWait(self):
        """Stop the AI from automatically moving"""
        self.onOutgoingData('force\n')
        
    def sendMovePrompt(self):
        """Get the AI to move for the current player"""
        self.onOutgoingData('go\n')

    def sendQuit(self):
        """Quit the engine"""
        # Send 'quit' starting with a newline in case there are some characters already sent
        self.onOutgoingData('\nquit\n')
    
    def registerIncomingData(self, data):
        """
        """
        self.__buffer += data
        self.__parseData()
        
    # Private methods
        
    def __parseData(self):
        while True:
            index = self.__buffer.find(self.NEWLINE)
            if index < 0:
                return
            
            line = self.__buffer[:index]
            self.__buffer = self.__buffer[index+1:]

            self.__parseLine(line)
    
    def __parseLine(self, line):
        for prefix in self.MOVE_PREFIXS:
            if line.startswith(prefix):
                move = line[len(prefix):]
                self.logText(line + '\n', 'move')
                self.onMove(move.strip())
                return
            
        if line.startswith(self.INVALID_MOVE_PREFIX):
            self.onIllegalMove(line[len(self.INVALID_MOVE_PREFIX):])
    
        elif line.startswith(self.RESIGN_PREFIX):
            self.onResign()
            
        elif line.startswith(self.DRAW_PREFIX):
            print 'AI calls a draw'

        else:
            self.onUnknownLine(line)
            
        self.logText(line + '\n', 'input')

import select

class CECPConnection(CECPProtocol):
    """
    """
    
    __pipeFromEngine = None
    __pipeToEngine = None
    
    # PID of the subprocess containing the engine
    __pid = None
    
    __connected = False
    
    def __init__(self, executable, arguments = []):
        """Create a connection to a CECP AI.
        
        'executable' is the name of the CECP executable (string).
        'arguments' are optional command-line arguments to pass to the engine (list of strings).
        """
        # Check this is a valid executable
        # TODO
        
        # Create pipes for communication with AI process
        self.__pipeToEngine = os.pipe()
        self.__pipeFromEngine = os.pipe()

        # Fork sub-process for engine
        self.__pid = os.fork()

        # This is the forked process, replace it with the engine
        if self.__pid == 0:
            self.__startEngine(executable, arguments)
        self.__connected = True
            
        # Close the ends of the pipe we are not using
        os.close(self.__pipeFromEngine[1]);
        os.close(self.__pipeToEngine[0]);
        
        # Start protocol
        CECPProtocol.__init__(self)
        
    # Methods to extend
    
    def logText(self, text, style):
        """FIXME: define style
        """
        pass
    
    # Public methods

    def read(self):
        """
        """
        if self.__connected is False:
            return
        
        while True:
            (rlist, _, _) = select.select([self.__pipeFromEngine[0]], [], [], 0)
            if len(rlist) == 0:
                return
            data = os.read(self.__pipeFromEngine[0], 256)
            if data == '':
                return
            self.registerIncomingData(data)
  
    def __startEngine(self, executable, arguments):
        """
        """
        # Connect stdin and stdout to the pipes to the main process
        os.dup2(self.__pipeFromEngine[1], sys.stdout.fileno())
        os.dup2(self.__pipeToEngine[0], sys.stdin.fileno())
        
        # Close the ends of the pipe we are not using
        os.close(self.__pipeFromEngine[0]);
        os.close(self.__pipeToEngine[1]);
        
        # Make the process nice so it doesn't hog the CPU
        os.nice(15)

        # Start the AI
        try:
            os.execv(executable, arguments)
        except OSError:
            select.select([], [], [], None)
            os._exit(os.EX_UNAVAILABLE)
        
    # Callbacks
    
    def onAIMove(self, move):
        """Called when the AI makes a move.
        
        'move' is the move the AI made (string).
        """
        print 'AI moves: ' + move

    # Public methods
    
    def requestMove(self):
        """Request the AI moves for the current player"""
        # Prompt the AI to move
        self.sendMovePrompt()
        
    def reportMove(self, move):
        """Report the move the current player has made.
        
        'move' is the move to report (string).
        
        Only report moves not made by the AI.
        """
        # Stop the AI from automatically moving
        self.sendWait()

        # Report the move
        self.sendMove(move)
        
    def quit(self):
        """Disconnect the AI"""
        self.__connected = False
        
        # Wait for the pipe to close
        # There must be a better way of doing this!
        count = 0
        while True:
            select.select([], [], [], 0.1)
            try:
                os.write(self.__pipeToEngine[1], '\nquit\n')
            except OSError:
                return
            count += 1
            if count > 5:
                break
        
        print 'Killing AI ' + str(self.__pid)
        os.kill(self.__pid, 9)

    # Private methods
        
    def onOutgoingData(self, data):
        """Called by CECPProtocol"""
        self.logText(data, 'output')
        try:
            os.write(self.__pipeToEngine[1], data)
        except OSError, e:
            print 'Failed to write to engine: ' + str(e)
    
    def onUnknownLine(self, line):
        """Called by CECPProtocol"""
        pass#print 'Unknown CECP line: ' + line

    def onMove(self, move):
        """Called by CECPProtocol"""
        self.onAIMove(move)
    
    def onIllegalMove(self, move):
        """Called by CECPProtocol"""
        print 'CECP illegal move: ' + move

if __name__ == '__main__':
    c = CECPConnection('gnuchess')
    while True:
        c.read()
