"""
"""

__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

import chess.board
import chess.san

class ChessMove:
    """
    """
    # The move number (game starts at 0)
    number     = 0
    
    # The player and piece that moved
    player     = None
    piece      = None

    # The start and end position of the move
    start      = None
    end        = None
    
    # The move in CAN and SAN format
    canMove    = ''
    sanMove    = ''

class ChessPlayer:
    """
    """
    # The name of the player
    __name = None
    __type = None
    
    # The game this player is in
    __game = None

    # Flag to show if this player is able to move
    __readyToMove = False
    
    def __init__(self, name):
        """Constructor for a chess player.

        'name' is the name of the player.
        """
        self.__name = str(name)

    # Methods to extend

    def onPieceMoved(self, piece, start, end):
        """Called when a chess piece is moved.
        
        'piece' is the piece that has been moved (chess.board.ChessPiece).
        'start' is the location the piece has moved from (tuple, (file,rank)) or None if the piece has been created.
        'end' is the location the piece has moved to (tuple, (file,rank)) or None if the piece has been deleted.
        """
        pass
    
    def onPlayerMoved(self, player, move):
        """Called when a player has moved.
        
        'player' is the player that has moved (ChessPlayer).
        'move' is the record for this move (ChessMove).
        """
        pass
    
    def onGameEnded(self, winningPlayer = None):
        """Called when a chess game has ended.
        
        'winningPlayer' is the player that won or None if the game was a draw.
        """
        pass

    # Public methods
        
    def getName(self):
        """Get the name of this player.
        
        Returns the player name (string).
        """
        return self.__name

    def readyToMove(self):
        """
        """
        return self.__readyToMove
    
    def canMove(self, move, promotionType = chess.board.QUEEN):
        """
        """
        return self.__game.canMove(self, move, promotionType)

    def move(self, move):
        """
        """
        self.__game.move(self, move)
        
    # Private methods
    
    def _setGame(self, game):
        """
        """
        self.__game = game
        
    def _setReadyToMove(self, readyToMove):
        self.__readyToMove = readyToMove
        if readyToMove is True:
            self.readyToMove()

class ChessGameBoard(chess.board.ChessBoard):
    """
    """
    
    # Reference to the game
    __game = None
    
    def __init__(self, game):
        """
        """
        self.__game = game
        chess.board.ChessBoard.__init__(self)

    def onPieceMoved(self, piece, start, end):
        """Called by chess.board.ChessBoard"""
        self.__game._onPieceMoved(piece, start, end)

class ChessGameSANConverter(chess.san.SANConverter):
    """
    """
        
    __colourToSAN = {chess.board.WHITE: chess.san.SANConverter.WHITE,
                     chess.board.BLACK: chess.san.SANConverter.BLACK}
    __sanToColour = {}
    for (a, b) in __colourToSAN.iteritems():
        __sanToColour[b] = a
        
    __typeToSAN = {chess.board.PAWN:   chess.san.SANConverter.PAWN,
                   chess.board.KNIGHT: chess.san.SANConverter.KNIGHT,
                   chess.board.BISHOP: chess.san.SANConverter.BISHOP,
                   chess.board.ROOK:   chess.san.SANConverter.ROOK,
                   chess.board.QUEEN:  chess.san.SANConverter.QUEEN,
                   chess.board.KING:   chess.san.SANConverter.KING}
    __sanToType = {}
    for (a, b) in __typeToSAN.iteritems():
        __sanToType[b] = a
        
    __resultToSAN = {chess.board.MOVE_RESULT_ILLEGAL: False,
                     chess.board.MOVE_RESULT_ALLOWED: True,
                     chess.board.MOVE_RESULT_OPPONENT_CHECK: chess.san.SANConverter.CHECK,
                     chess.board.MOVE_RESULT_OPPONENT_CHECKMATE: chess.san.SANConverter.CHECKMATE}
        
    __board = None
        
    def __init__(self, board):
        self.__board = board
        chess.san.SANConverter.__init__(self)
    
    def decodeSAN(self, colour, move):
        (start, end, result, promotionType) = chess.san.SANConverter.decodeSAN(self, self.__colourToSAN[colour], move)
        return (start, end, self.__sanToType[promotionType])
    
    def encodeSAN(self, start, end, promotionType):
        return chess.san.SANConverter.encodeSAN(self, start, end, self.__typeToSAN[promotionType])

    def getPiece(self, file, rank):
        """Called by chess.san.SANConverter"""
        piece = self.__board.getPiece((file, rank))
        if piece is None:
            return None
        return (self.__colourToSAN[piece.getColour()], self.__typeToSAN[piece.getType()])
    
    def testMove(self, colour, start, end, promotionType, allowSuicide = False):
        """Called by chess.san.SANConverter"""
        moveResult = self.__board.testMove(self.__sanToColour[colour], ((start, end)), self.__sanToType[promotionType], allowSuicide)
        
        return self.__resultToSAN[moveResult]
    
class ChessGameLANConverter:
    """
    """
    __board = None
    
    __fileToLAN = {0:'a', 1:'b', 2:'c', 3:'d', 4:'e', 5:'f', 6:'g', 7:'h'}
    __rankToLAN = {0:'1', 1:'2', 2:'3', 3:'4', 4:'5', 5:'6', 6:'7', 7:'8'}
    __lanToFile = {'a':0, 'b':1, 'c':2, 'd':3, 'e':4, 'f':5, 'g':6, 'h':7}
    __lanToRank = {'1':0, '2':1, '3':2, '4':3, '5':4, '6':5, '7':6, '8':7}
    
    __typeToLAN = {chess.board.PAWN:   'Pp',
                   chess.board.KNIGHT: 'Nn',
                   chess.board.BISHOP: 'Bb',
                   chess.board.ROOK:   'Rr',
                   chess.board.QUEEN:  'Qq',
                   chess.board.KING:   'Kk'}
    __lanToType = {}
    for (a, b) in __typeToLAN.iteritems():
        for c in b:
            __lanToType[c] = a
    
    def __init__(self, board):
        """
        """
        self.__board = board
        
    def decode(self, colour, move):
        """
        """
        promotionType = chess.board.QUEEN
        
        if colour is chess.board.WHITE:
            baseRank = 0
        else:
            baseRank = 7
        if move == 'O-O':
            return (chess.board.KING, (4, baseRank), (6, baseRank), False, promotionType)
        elif move == 'O-O-O':
            return (chess.board.KING, (4, baseRank), (2, baseRank), False, promotionType)

        if len(move) < 1:
            raise TypeError('Too short')
        try:
            pieceType = self.__lanToType[move[0]]
        except KeyError:
            pieceType = None
        else:
            move = move[1:]
        
        if len(move) < 2:
            raise TypeError('Too short')
        try:
            start = (self.__lanToFile[move[0]], self.__lanToRank[move[1]])
        except KeyError:
            raise TypeError('Invalid format')
        move = move[2:]
        
        if len(move) < 1:
            raise TypeError('Too short')
        if move[0] == '-':
            isTake = False
            move = move[1:]
        elif move[0] == 'x':
            isTake = True
            move = move[1:]
        else:
            isTake = False

        if len(move) < 2:
            raise TypeError('Too short')
        try:
            end = (self.__lanToFile[move[0]], self.__lanToRank[move[1]])
        except KeyError:
            raise TypeError('Invalid format')
        move = move[2:]

        # Look for promotion type
        if len(move) > 0:
            if move[0] == '=':
                if len(move) < 2:
                    raise TypeError('Too short')
                try:
                    promotionType = self.__lanToType[move[1]]
                except KeyError:
                    raise TypeError('Unknown promotion type')
                move = move[2:]
            else:
                try:
                    promotionType = self.__lanToType[move[0]]
                except KeyError:
                    pass
                else:
                    move = move[1:]
            
        if len(move) > 0:
            if move[0] == '+' or move[0] == '#':
                move = move[1:]

        if len(move) != 0:
            raise TypeError('Extra characters')
        return (pieceType, start, end, isTake, promotionType)
    
    def encode(self, start, end, promotionType):
        """Encode a move in Long Algebraic Notation (LAN).
        
        'start' is the location of the piece being moved (tuple (file,rank)).
        'end' is the location of the piece is being moved to (tuple (file,rank)).
        'promotionType' is the piece type pawns are promoted to.
        
        The move is not checked if it is valid.
        """
        # Get the piece name
        piece = self.__board.getPiece(start)
        move = self.__typeToLAN[piece.getType()][0]
        
        # Start position
        move += self.__fileToLAN[start[0]][0] + self.__rankToLAN[start[1]][0]
        
        # Check if this move takes another piece
        target = self.__board.getPiece(end)
        if target is None:
            move += '-'
        else:
            assert(target.getColour() != piece.getColour())
            move += 'x'
        
        # End position
        move += self.__fileToLAN[end[0]][0] + self.__rankToLAN[end[1]][0]
        
        # Provide promotion type if required
        if piece is None:
            return move
        
        # If the moved piece is a pawn and it has moved to an edge of the board then promote it.
        # Note: Since pawns can't move backwards we don't have to check colour.
        if piece.getType() is chess.board.PAWN and (end[1] == 0 or end[1] == 7):
            move += '=' + self.__typeToLAN[promotionType][0]
    
        return move
    
class ChessGameCANConverter:
    """
    """
    
    __board = None
    
    __fileToCAN = {0:'a', 1:'b', 2:'c', 3:'d', 4:'e', 5:'f', 6:'g', 7:'h'}
    __rankToCAN = {0:'1', 1:'2', 2:'3', 3:'4', 4:'5', 5:'6', 6:'7', 7:'8'}
    __canToFile = {'a':0, 'b':1, 'c':2, 'd':3, 'e':4, 'f':5, 'g':6, 'h':7}
    __canToRank = {'1':0, '2':1, '3':2, '4':3, '5':4, '6':5, '7':6, '8':7}
    
    __typeToCAN = {chess.board.PAWN:   'p',
                   chess.board.KNIGHT: 'n',
                   chess.board.BISHOP: 'b',
                   chess.board.ROOK:   'r',
                   chess.board.QUEEN:  'q',
                   chess.board.KING:   'k'}
    __canToType = {}
    for (a, b) in __typeToCAN.iteritems():
        __canToType[b] = a
    
    def __init__(self, board):
        """
        """
        self.__board = board
            
    def encode(self, start, end, promotionType):
        """Encode a move in Co-ordinate Algebraic Notation (CAN).
        
        'start' is the location of the piece being moved (tuple (file,rank)).
        'end' is the location of the piece is being moved to (tuple (file,rank)).
        'promotionType' is the piece type pawns are promoted to.
        
        The move is not checked if it is valid.
        """
        move = self.__fileToCAN[start[0]] + self.__rankToCAN[start[1]] + self.__fileToCAN[end[0]] + self.__rankToCAN[end[1]]
        
        # Provide promotion type if required
        piece = self.__board.getPiece(start)
        if piece is None:
            return move
        
        # If the moved piece is a pawn and it has moved to an edge of the board then promote it.
        # Note: Since pawns can't move backwards we don't have to check colour.
        if piece.getType() is chess.board.PAWN and (end[1] == 0 or end[1] == 7):
            move += self.__typeToCAN[promotionType]
        
        return move

    def decode(self, move):
        """Decode a move in Co-ordinate Algebraic Notation (CAN).
        
        'move' is the CAN move to decode (string).
        
        Returns a tuple containing (start, end, promotionType).
        If the move is invalid then a TypeError exception is raised.
        """
        if len(move) < 4:
            raise TypeError('Move ' + repr(move) + ' is too short for CAN')
        elif len(move) > 5:
            raise TypeError('Move ' + repr(move) + ' is too long for CAN')
            
        try:
            start = (self.__canToFile[move[0]], self.__canToRank[move[1]])
            end = (self.__canToFile[move[2]], self.__canToRank[move[3]])
        except KeyError:
            raise TypeError('Move ' + repr(move) + ' contains invalid co-ordinates')
        
        if len(move) == 4:
            return (start, end, chess.board.QUEEN)
        
        try:
            promotionType = self.__canToType[move[4]]
        except KeyError:
            raise TypeError('Move ' + repr(move) + ' contains invalid promotion type')
        
        return (start, end, promotionType)

class ChessGame:
    """
    """    
    # The players and spectators in the game
    __players = None
    __whitePlayer = None
    __blackPlayer = None
    __spectators = None
    
    # The board to move on
    __board = None
    
    # SAN, CAN and LAN en/decoders
    __sanConverter = None
    __canConverter = None
    __lanConverter = None
    
    # The game state (started and player to move)
    __started = False
    __currentPlayer = None

    # Flag to show if calling a player and the queued up moves
    __notifyingPlayer = False
    __queuedMoves = None
    
    __moves = None
    
    def __init__(self):
        """Game constructor"""
        self.__players = []
        self.__spectators = []
        self.__board = ChessGameBoard(self)
        self.__sanConverter = ChessGameSANConverter(self.__board)
        self.__canConverter = ChessGameCANConverter(self.__board)
        self.__lanConverter = ChessGameLANConverter(self.__board)
        self.__moves = []
        
    def getAlivePieces(self, moveNumber = -1):
        """Get the alive pieces on the board.
        
        'moveNumber' is the move to get the pieces from (integer).
        
        Returns a dictionary of the alive pieces (board.ChessPiece) keyed by location.
        Raises an IndexError exception if moveNumber is invalid.
        """
        return self.__board.getAlivePieces(moveNumber)
    
    def getDeadPieces(self, moveNumber = -1):
        """Get the dead pieces from the game.
        
        'moveNumber' is the move to get the pieces from (integer).
        
        Returns a list of the pieces (board.ChessPiece) in the order they were killed.
        Raises an IndexError exception if moveNumber is invalid.
        """
        return self.__board.getDeadPieces(moveNumber)

    def setWhite(self, player):
        """Set the white player in the game.
        
        'player' is the player to use as white.
        
        If the game has started or there is a white player an exception is thrown.
        """
        assert(self.__started is False)
        assert(self.__whitePlayer is None)
        self.__whitePlayer = player
        self.__connectPlayer(player)
        
    def getWhite(self):
        """Returns the current white player (player.Player)"""
        return self.__whitePlayer
    
    def setBlack(self, player):
        """Set the black player in the game.
        
        'player' is the player to use as black.
        
        If the game has started or there is a black player an exception is thrown.
        """
        assert(self.__started is False)
        assert(self.__blackPlayer is None)
        self.__blackPlayer = player
        self.__connectPlayer(player)
        
    def getBlack(self):
        """Returns the current white player (player.Player)"""
        return self.__blackPlayer
    
    def addSpectator(self, player):
        """Add a spectator to the game.
        
        'player' is the player spectating.
        
        This can be called after the game has started.
        """
        self.__spectators.append(player)
        self.__connectPlayer(player)
        
    def start(self, moves = []):
        """Start the game.
        
        'moves' is a list of moves to start with.
        
        If there is no white or black player then an exception is raised.
        """
        assert(self.__whitePlayer is not None and self.__blackPlayer is not None)
        
        # Disabled for now
        #import network
        #self.x = network.GameReporter('Test Game', 12345)
        #print 'Reporting'
        
        # Set initial state
        self.__queuedMoves = []
        self.__currentPlayer = self.__whitePlayer
        
        # Load starting moves
        try:
            for move in moves:
                self.move(self.__currentPlayer, move)
        except chess.san.Error, e:
            print e
            
        self.__started = True
        
        # Get the next player to move
        self.__currentPlayer._setReadyToMove(True)
        
    def getSquareOwner(self, coord):
        """TODO
        """
        piece = self.__board.getPiece(coord)
        if piece is None:
            return None
        
        colour = piece.getColour()
        if colour is chess.board.WHITE:
            return self.__whitePlayer
        elif colour is chess.board.BLACK:
            return self.__blackPlayer
        else:
            return None
        
    def canMove(self, player, move, promotionType):
        """Test if a player can move.
        
        'player' is the player making the move.
        'move' is the move to make (TODO).
        'promotionType' is the piece type to promote pawns to. FIXME: Make this a property of the player
        
        Return True if can move, otherwise False.
        """
        if player is not self.__currentPlayer:
            return False
        
        if player is self.__whitePlayer:
            colour = chess.board.WHITE
        elif player is self.__blackPlayer:
            colour = chess.board.BLACK
        else:
            assert(False)

        # FIXME: Convert move from SAN etc
        
        moveResult = self.__board.testMove(colour, move, promotionType = promotionType)
         
        return moveResult is not chess.board.MOVE_RESULT_ILLEGAL
    
    def move(self, player, move):
        """Get a player to make a move.
        
        'player' is the player making the move.
        'move' is the move to make. It can be of the form:
               A coordinate move in the form ((file0, rank0), (file1, rank1), promotionType) ((int, int), (int, int), chess.board.PIECE_TYPE) or
               A SAN move (string).
        """
        self.__queuedMoves.append((player, move))
        
        # Don't process if still finishing the last move
        if self.__notifyingPlayer:
            return
        
        while True:
            try:
                (movingPlayer, move) = self.__queuedMoves.pop(0)
            except IndexError:
                return

            if movingPlayer is not self.__currentPlayer:
                print 'Player attempted to move out of turn'
            else:
                self.__notifyingPlayer = True
                self._move(movingPlayer, move)
                self.__notifyingPlayer = False
            
    def _move(self, player, move):
        """
        """        
        if self.__currentPlayer is self.__whitePlayer:
            nextPlayer = self.__blackPlayer
            colour = chess.board.WHITE
        else:
            nextPlayer = self.__whitePlayer
            colour = chess.board.BLACK
                
        # If move is SAN process it as such
        if type(move) is str:
            try:
                (start, end, promotionType) = self.__canConverter.decode(move)
            except TypeError:
                try:
                    (_, start, end, _, promotionType) = self.__lanConverter.decode(colour, move)
                except TypeError:
                    (start, end, promotionType) = self.__sanConverter.decodeSAN(colour, move)
        elif type(move) is tuple and len(move) == 3:
            start = move[0]
            end = move[1]
            promotionType = move[2]
        else:
            raise TypeError() # FIXME

        # FIXME: Don't apply if the move is invalid
        sanMove = self.__sanConverter.encodeSAN(start, end, promotionType)
        canMove = self.__canConverter.encode(start, end, promotionType)
        moveResult = self.__board.movePiece(colour, (start, end), promotionType)
            
        if moveResult is chess.board.MOVE_RESULT_ILLEGAL:
            print 'Illegal move: ' + str(move)
            return
            
        move = ChessMove()
        if len(self.__moves) == 0:
            move.number = 1
        else:
            move.number = self.__moves[-1].number + 1
        move.player  = self.__currentPlayer
        move.start   = start
        move.end     = end
        move.canMove = canMove
        move.sanMove = sanMove
        self.__moves.append(move)

        # This player has now moved
        self.__currentPlayer._setReadyToMove(False)
            
        # Inform other players of the result
        for player in self.__players:
            player.onPlayerMoved(self.__currentPlayer, move)
            
        # Notify the next player they can move
        self.__currentPlayer = nextPlayer
        if self.__started is True:
            nextPlayer._setReadyToMove(True)

    def getMoves(self):
        """
        """
        return self.__moves[:]
    
    def end(self):
        """End the game"""
        # Inform players
        for player in self.__players:
            player.onGameEnded()

    # Private methods:

    def __connectPlayer(self, player):
        """Add a player into the game.
        
        'player' is the player to add.
        
        The player will be notified of the current state of the board.
        """
        self.__players.append(player)
        player._setGame(self)
        
        # Notify the player of the current state
        # FIXME: Make the board iteratable...
        for file in [0, 1, 2, 3, 4, 5, 6, 7]:
            for rank in [0, 1, 2, 3, 4, 5, 6, 7]:
                coord = (file, rank)
                piece = self.__board.getPiece(coord)
                if piece is None:
                    continue

                # These are moves from nowhere to their current location
                player.onPieceMoved(piece, None, coord)
                
    def _onPieceMoved(self, piece, start, end):
        """Called by the chess board"""
        
        # Notify all players of creations and deletions
        # NOTE: Normal moves are done above since the SAN moves are calculated before the move...
        # FIXME: Change this so the SAN moves are done afterwards...
        for player in self.__players:
            player.onPieceMoved(piece, start, end)
            
class NetworkChessGame(ChessGame):
    """
    """
    
    def move(self, player, move):
        """Get a player to make a move.
        
        'player' is the player making the move.
        'move' is the move to make. It can be of the form:
               A coordinate move in the form ((file0, rank0), (file1, rank1), promotionType) ((int, int), (int, int), chess.board.PIECE_TYPE) or
               A SAN move (string).
        """
        # Send to the server
        
            
if __name__ == '__main__':
    game = ChessGame()
    
    import pgn
    
    p = pgn.PGN('black.pgn')
    g = p.getGame(0)

    class PGNPlayer(ChessPlayer):
        __moveNumber = 1
        
        __isWhite = True
        
        def __init__(self, isWhite):
            self.__isWhite = isWhite
        
        def readyToMove(self):
            if self.__isWhite:
                move = g.getWhiteMove(self.__moveNumber)
            else:
                move = g.getBlackMove(self.__moveNumber)
            self.__moveNumber += 1
            self.move(move)
            
    white = PGNPlayer(True)
    black = PGNPlayer(False)
    
    game.setWhite(white)
    game.setBlack(black)
    
    game.start()
    