__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

__all__ = ['GtkView', 'GtkUI']

# TODO: Extend base UI classes?

import os
import sys
import traceback
import time
import gettext
import ConfigParser

import gobject
import gtk
import gtk.glade
import gtk.gdk
import gtk.gtkgl
import pango

import glchess.ui
import dialogs

# TEMP
from OpenGL.GL import *

# Stop PyGTK from catching exceptions
os.environ['PYGTK_FATAL_EXCEPTIONS'] = '1'

UI_DATA_DIR = 'lib/glchess/gtkui/'
if not os.path.isdir(UI_DATA_DIR):
    UI_DATA_DIR = '/usr/share/games/glchess/gui/'

class GtkViewArea(gtk.DrawingArea, gtk.gtkgl.Widget):
    """Custom widget to render an OpenGL scene"""
    # The view this widget is rendering
    view = None
    
    renderGL = False

    # Pixmaps to use for double buffering
    pixmap = None
    dynamicPixmap = None
    
    # TODO...
    __glDrawable = None
    
    def __init__(self, view):
        """
        """
        gtk.DrawingArea.__init__(self)
        gtk.gtkgl.Widget.__init__(self)
        
        self.view = view

        # Allow notification of button presses
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.BUTTON_MOTION_MASK)
        
        # Make openGL drawable
        self.set_gl_capability(self.view.ui.notebook.glConfig)# FIXME:, share_list=glContext)

        # Connect signals
        self.connect('realize', self.__init)
        self.connect('configure_event', self.__configure)
        self.connect('expose_event', self.__expose)
        self.connect('button_press_event', self.__button_press)
        self.connect('button_release_event', self.__button_release)
        
    # Public methods
        
    def redraw(self):
        """Request this widget is redrawn"""
        #FIXME: Check this is valid
        self.window.invalidate_rect(self.allocation, False)
        
    def setRenderGL(self, renderGL):
        """Enable OpenGL rendering"""
        if self.renderGL == renderGL:
            return
        self.renderGL = renderGL
        self.redraw()
    
    # Private methods

    def __startGL(self):
        """Get the OpenGL context"""
        if not self.renderGL:
            return

        assert(self.__glDrawable is None)
        
        # Obtain a reference to the OpenGL drawable
        # and rendering context.
        glDrawable = self.get_gl_drawable()
        glContext = self.get_gl_context()

        # OpenGL begin.
        if not glDrawable.gl_begin(glContext):
            return
        
        self.__glDrawable = glDrawable
        
        if not self.view.ui.openGLInfoPrinted:
            print 'Using OpenGL:'
            print 'VENDOR=' + glGetString(GL_VENDOR)
            print 'RENDERER=' + glGetString(GL_RENDERER)
            print 'VERSION=' + glGetString(GL_VERSION)
            print 'EXTENSIONS=' + glGetString(GL_EXTENSIONS)
            self.view.ui.openGLInfoPrinted = True
        
    def __endGL(self):
        """Free the OpenGL context"""
        if not self.renderGL:
            return
        
        assert(self.__glDrawable is not None)
        self.__glDrawable.gl_end()
        self.__glDrawable = None
        
    def __init(self, widget):
        """Gtk+ signal"""
        if self.view.feedback is not None:
            self.view.feedback.reshape(widget.allocation.width, widget.allocation.height)
        
    def __configure(self, widget, event):
        """Gtk+ signal"""
        self.pixmap = gtk.gdk.Pixmap(widget.window, event.width, event.height)
        self.dynamicPixmap = gtk.gdk.Pixmap(widget.window, event.width, event.height)
        self.__startGL()
        if self.view.feedback is not None:
            self.view.feedback.reshape(event.width, event.height)
        self.__endGL()

    def __expose(self, widget, event):
        """Gtk+ signal"""
        if self.renderGL:
            self.__startGL()
        
            # Get the scene rendered
            try:
                if self.view.feedback is not None:
                    self.view.feedback.renderGL()
            except GLerror, e:
                print 'Rendering Error: ' + str(e)
                traceback.print_exc(file = sys.stdout)

            # Paint this
            if self.__glDrawable.is_double_buffered():
                self.__glDrawable.swap_buffers()
            else:
                glFlush()

            self.__endGL()
            
        else:
            context = self.pixmap.cairo_create()
            if self.view.feedback is not None:
                self.view.feedback.renderCairoStatic(context)
            
            # Copy the background to render the dynamic elements on top
            self.dynamicPixmap.draw_drawable(widget.get_style().white_gc, self.pixmap, 0, 0, 0, 0, -1, -1)
            context = self.dynamicPixmap.cairo_create()
        
            # Set a clip region for the expose event
            context.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
            context.clip()
           
            # Render the dynamic elements
            if self.view.feedback is not None:
                self.view.feedback.renderCairoDynamic(context)
                
            # Draw the window
            widget.window.draw_drawable(widget.get_style().white_gc, self.dynamicPixmap,
                                        event.area.x, event.area.y,
                                        event.area.x, event.area.y, event.area.width, event.area.height)

    def __button_press(self, widget, event):
        """Gtk+ signal"""
        self.__startGL()
        if self.view.feedback is not None:
            self.view.feedback.select(event.x, event.y)
        self.__endGL()
        
    def __button_release(self, widget, event):
        """Gtk+ signal"""
        self.__startGL()
        if self.view.feedback is not None:
            self.view.feedback.deselect(event.x, event.y)
        self.__endGL()
        
class GtkView(glchess.ui.ViewController):
    """
    """
    # The UI this view belongs to
    ui            = None
    
    # The widget to render the scene to
    widget        = None
    
    # A Gtk+ tree model to store the move history
    moveModel    = None
    selectedMove = -1
    
    def __init__(self, ui, feedback, isActive = True):
        """Constructor for a view.
        
        'feedback' is the feedback object for this view (extends ui.ViewFeedback).
        'isActive' is a flag showing if this view can be controlled by the user (True) or not (False).
        """
        self.ui = ui
        self.feedback = feedback
        self.isActive = isActive
        self.widget = GtkViewArea(self)

        # Make a model for navigation
        model = gtk.ListStore(int, str)
        iter = model.append()
        model.set(iter, 0, 0, 1, gettext.gettext('Game Start'))
        self.moveModel = model

        self.widget.show_all()
    
    # Extended methods
    
    def render(self):
        """Extends glchess.ui.ViewController"""
        self.widget.redraw()
        
    def addMove(self, move):
        """Extends glchess.ui.ViewController"""
        # FIXME: Make a '@ui' player who watches for these itself?
        iter = self.moveModel.append()
        string = '%2i. ' % ((move.number - 1) / 2 + 1)
        if move.number % 2 == 0:
            string += '... '
        string += move.sanMove
        self.moveModel.set(iter, 0, move.number, 1, string)
        
        # If is the current view and tracking the game select this
        if self.selectedMove == -1:
            self.ui._updateViewButtons()
    
    def close(self):
        """Extends glchess.ui.ViewController"""
        self.ui._removeView(self)
    
    # Public methods

    def _getModel(self):
        """
        """
        return (self.moveModel, self.selectedMove)
        
    def _setMoveNumber(self, moveNumber):
        """Set the move number this view requests.
        
        'moveNumber' is the move number to use (integer).
        """
        self.selectedMove = moveNumber
        if self.feedback is not None:
            self.feedback.setMoveNumber(moveNumber)
        
class GtkGameNotebook(gtk.Notebook):
    """
    """
    glConfig          = None

    defaultView       = None
    viewsByWidget     = None
    
    def __init__(self, ui):
        """
        """
        self.ui = ui
        self.viewsByWidget = {}
        
        gtk.Notebook.__init__(self)
        self.set_show_border(False)
        
        # Make the tabs scrollable so the area is not resized
        self.set_scrollable(True)
        
        # Hide tabs until more than one view is present
        self.set_show_tabs(False)
        
        try:
            import gtk.gtkgl
        except ImportError:
            self.glConfig = None
        else:
            # Configure openGL
            display_mode = (gtk.gdkgl.MODE_RGB | gtk.gdkgl.MODE_DEPTH | gtk.gdkgl.MODE_DOUBLE)
            try:
                self.glConfig = gtk.gdkgl.Config(mode=display_mode)
            except gtk.gdkgl.NoMatches:
                display_mode &= ~gtk.gdkgl.MODE_DOUBLE
                self.glConfig = gtk.gdkgl.Config(mode=display_mode)
        
    def setDefault(self, feedback):
        """
        """
        assert(self.defaultView is None)
        self.defaultView = GtkView(self.ui, feedback)
        page = self.append_page(self.defaultView.widget)
        self.set_current_page(page)
        
        self.__updateTabVisibleState()
        
        return self.defaultView

    def addView(self, title, feedback):
        """
        """
        view = GtkView(self.ui, feedback)
        self.viewsByWidget[view.widget] = view
        page = self.append_page(view.widget)
        self.set_tab_label_text(view.widget, title)
        self.set_current_page(page)
        
        self.__updateTabVisibleState()
        
        return view
    
    def getView(self, pageNumber = None):
        """Get the view at a given page number.
        
        'pageNumber' is the page to check or None to get the selected page.
        
        Return the view (GtkView) on this page or None if no view here.
        """
        if pageNumber is None:
            # If splashscreen present then there is no view
            if len(self.viewsByWidget) > 0:
                num = self.get_current_page()
                if num < 0:
                    return None
                widget = self.get_nth_page(num)
            else:
                return None
        else:
            widget = self.get_nth_page(pageNumber)
            
        return self.viewsByWidget[widget]
    
    def removeView(self, view):
        """Remove a view from the notebook.
        
        'view' is the view to remove.
        """
        self.remove_page(self.page_num(view.widget))
        self.viewsByWidget.pop(view.widget)
        self.__updateTabVisibleState()

    def __updateTabVisibleState(self):
        """
        """
        # Only show tabs if there is more than one game
        self.set_show_tabs(len(self.viewsByWidget) > 1)
        
        # Show/hide the default view
        if len(self.viewsByWidget) == 0:
            self.defaultView.widget.show()
        else:
            self.defaultView.widget.hide()
        
class AIWindow:
    """
    """
    
    notebook    = None
    defaultPage = None
    
    # We keep track of the number of pages as there is a bug 
    # in GtkNotebook (Gnome bug #331785).
    pageCount   = 0
    
    def __init__(self, notebook):
        """
        """
        self.notebook = notebook
        self.defaultPage = notebook.get_nth_page(0)
    
    def addView(self, title, executable, description):
        """
        """
        # Hide the default page
        self.defaultPage.hide()
        self.notebook.set_show_tabs(True)
        
        self.pageCount += 1
        return AIView(self, title, executable, description)

class AIView(gtk.glade.XML):
    """
    """
    
    def __init__(self, window, title, executable, description):
        """
        """
        self.window = window
        gtk.glade.XML.__init__(self, UI_DATA_DIR + 'ai.glade', 'ai_table', domain = 'glchess')
        self.get_widget('executable_label').set_text(executable)
        self.get_widget('game_label').set_text(description)

        # Add into the notebook
        self.root = self.get_widget('ai_table')
        notebook = window.notebook
        notebook.append_page(self.root, gtk.Label(title))
                
        # Create styles for the buffer
        buffer = self.get_widget('comms_textview').get_buffer()
        buffer.create_tag('input', family='Monospace')
        buffer.create_tag('output', family='Monospace', weight = pango.WEIGHT_BOLD)
        buffer.create_tag('move', family='Monospace', foreground = "red")
        
    def addText(self, text, style):
        """FIXME: Define style
        """
        buffer = self.get_widget('comms_textview').get_buffer()
        buffer.insert_with_tags_by_name(buffer.get_end_iter(), text, style)
        
    def close(self):
        """
        """
        self.window.pageCount -= 1
        self.window.notebook.remove_page(self.window.notebook.page_num(self.root))
        
        # Show the default page
        if self.window.pageCount == 0:
            self.window.defaultPage.show()
            self.window.notebook.set_show_tabs(False)

class GtkUI(glchess.ui.UI):
    """
    """
    # The Gtk+ GUI
    _gui              = None
    
    # The time stored for animation
    __lastTime        = None

    # The notebook containing games
    notebook        = None
    
    # The Gtk+ list model of the available player types
    __playerModel     = None
    
    # The about dialog open
    __aboutDialog     = None
    
    # Dictionary of save game dialogs keyed by view
    __saveGameDialogs = None
    
    # Dictionary of configuration options
    __config          = None
    __applyingConfig  = False
    
    __renderGL        = False
    openGLInfoPrinted = False

    # TODO
    __joinGameDialogs = None
    __networkGames    = None
    
    def __init__(self):
        """Constructor for a GTK+ glChess GUI"""
        self.__networkGames = {}
        self.__saveGameDialogs = {}
        self.__joinGameDialogs = []
        
        self._gui = gtk.glade.XML(UI_DATA_DIR + 'glchess.glade', domain = 'glchess')
        self._gui.signal_autoconnect(self)

        # Make a notebook for the games
        self.notebook = GtkGameNotebook(self)
        self.notebook.connect_after('switch-page', self._on_view_changed)
        self.__getWidget('game_viewport').add(self.notebook)
        self.notebook.show_all()
        
        # Create the model for the player types
        self.__playerModel = gtk.ListStore(gobject.TYPE_PYOBJECT, gtk.gdk.Pixbuf, str)
        iconTheme = gtk.icon_theme_get_default()
        icon = iconTheme.load_icon('stock_people', 24, gtk.ICON_LOOKUP_USE_BUILTIN)
        iter = self.__playerModel.append()
        self.__playerModel.set(iter, 0, None, 1, icon, 2, 'Human')
        # FIXME: Add spectators for network games
        
        self.__aiWindow = AIWindow(self._gui.get_widget('ai_notebook'))
        
        combo = self.__getWidget('history_combo')
        cell = gtk.CellRendererText()
        combo.pack_start(cell, False)
        combo.add_attribute(cell, 'text', 1)

        self.defaultViewController = self.notebook.setDefault(None)
        
        # Load UI preferences
        self.__config = {}
        self.__loadConfig()
        
        self.defaultViewController.widget.setRenderGL(self.__renderGL)

    # Public methods
    
    def addTimer(self, period, callback):
        """Add a timeout .
        
        'period' is the period for the timer in milliseconds (integer).
        'callback' is the method/function to call when the timer expires.
        
        The timer will be continuously called while 'callback' returns True.
        """
        gobject.timeout_add(period, callback)

    def addAIEngine(self, name):
        """Register an AI engine.
        
        'name' is the name of the engine.
        TODO: difficulty etc etc
        """
        iconTheme = gtk.icon_theme_get_default()
        icon = iconTheme.load_icon("stock_notebook", 24, gtk.ICON_LOOKUP_USE_BUILTIN)
        iter = self.__playerModel.append()
        self.__playerModel.set(iter, 0, name, 1, icon, 2, name)
        
    def setDefaultView(self, feedback):
        """Extends ui.UI"""
        self.defaultViewController.feedback = feedback
        return self.defaultViewController
        
    def addView(self, title, feedback):
        """Extends ui.UI"""
        view = self.notebook.addView(title, feedback)
        view.widget.setRenderGL(self.__renderGL)
        return view

    def addAIWindow(self, title, executable, description):
        """
        """
        return self.__aiWindow.addView(title, executable, description)

    def run(self):
        """Run the UI.
        
        This method will not return.
        """
        # Make an animation timer
        self.__lastTime = time.time()
        gobject.timeout_add(10, self.__animate)
        
        gtk.main()
        
    # Extended methods

    def reportError(self, title, error):
        """Extends glchess.ui.UI"""
        dialogs.GtkErrorDialog(title, error)
        
    def reportGameLoaded(self, gameName = None,
                         whiteName = None, blackName = None,
                         whiteAI = None, blackAI = None, moves = None):
        """Extends glchess.ui.UI"""
        dialogs.GtkNewGameDialog(self, self.__playerModel, gameName = gameName,
                                 whiteName = whiteName, whiteAI = whiteAI,
                                 blackName = blackName, blackAI = blackAI, moves = moves)
                                 
    def addNetworkGame(self, name, game):
        """Extends glchess.ui.UI"""
        self.__networkGames[game] = name
        
        # Update the open dialogs
        for dialog in self.__joinGameDialogs:
            dialog.addNetworkGame(name, game)
    
    def removeNetworkGame(self, game):
        """Extends glchess.ui.UI"""
        self.__networkGames.pop(game)

        # Update the open dialogs
        for dialog in self.__joinGameDialogs:
            dialog.removeNetworkGame(game)

    # Protected methods

    def _saveView(self, view, path):
        """
        """
        self.__saveGameDialogs.pop(view)
        if path is None:
            return
        
        if view.feedback is not None:
            view.feedback.save(path)
            
    def _removeView(self, view):
        """Remove a view from the UI.
        
        'view' is the view to remove.
        """
        self.notebook.removeView(view)
        self._updateViewButtons()

    # Private methods
    
    def __loadConfig(self):
        """
        """
        # Set defaults
        self.__config = {'show_toolbar': 'True',
                         'show_history': 'True',
                         'show_3d': 'False'}
        
        name = os.path.expanduser('~/.glchess/gtkui.ini')
        cp = ConfigParser.ConfigParser()
        cp.read(name)
        
        try:
            items = cp.items('ui')
        except ConfigParser.NoSectionError:
            self.__saveConfig()
        else:
            for (name, value) in items:
                self.__config[name] = value

        self.__applyConfig()

    def __setConfig(self, name, value):
        """
        """
        if self.__applyingConfig:
            return
        
        self.__config[name] = value
        self.__applyConfig()
        self.__saveConfig()
        
    def __saveConfig(self):
        """
        """
        name = os.path.expanduser('~/.glchess/gtkui.ini')
        try:
            f = file(name, 'w')
        except IOError, e:
            print 'Unable to save config: ' + str(e)
            return
        
        cp = ConfigParser.ConfigParser()
        cp.add_section('ui')
        for (name, value) in self.__config.iteritems():
            cp.set('ui', name, value)
        cp.write(f)
        
    def __applyConfig(self):
        """
        """
        # Stop recursion
        self.__applyingConfig = True
        
        # Show/hide the toolbar
        toolbar = self.__getWidget('toolbar')
        menu = self.__getWidget('menu_view_toolbar')
        if self.__config['show_toolbar'] == 'True':
            menu.set_active(True)
            toolbar.show()
        else:
            menu.set_active(False)
            toolbar.hide()
            
        # Show/hide the history
        box = self.__getWidget('navigation_box')
        menu = self.__getWidget('menu_view_history')
        if self.__config['show_history'] == 'True':
            menu.set_active(True)
            box.show()
        else:
            menu.set_active(False)
            box.hide()

        # Enable/disable OpenGL rendering
        self.__renderGL = self.__config['show_3d'] == 'True'
        menuItem = self.__getWidget('menu_view_3d')
        menuItem.set_active(self.__renderGL)
        self.notebook.defaultView.widget.setRenderGL(self.__renderGL)
        for view in self.notebook.viewsByWidget.itervalues():
            view.widget.setRenderGL(self.__renderGL)
            
        self.__applyingConfig = False
    
    def __animate(self):
        # Get the timestep, if it is less than zero or more than a second
        # then the system clock was probably changed.
        now = time.time()
        step = now - self.__lastTime
        if step < 0.0:
            step = 0.0
        elif step > 1.0:
            step = 1.0
        self.__lastTime = now
        
        # Animate!
        self.onAnimate(step)

        # Keep the timer running (deletes itself when returns False)
        return True

    def __getWidget(self, name):
        return self._gui.get_widget(name)
    
    def _on_show_toolbar_clicked(self, widget, data = None):
        """Gtk+ callback"""
        if widget.get_active():
            value = 'True'
        else:
            value = 'False'
        self.__setConfig('show_toolbar', value)

    def _on_show_history_clicked(self, widget, data = None):
        """Gtk+ callback"""
        if widget.get_active():
            value = 'True'
        else:
            value = 'False'
        self.__setConfig('show_history', value)

    def _on_toggle_3d_clicked(self, widget, data = None):
        """Gtk+ callback"""
        if widget.get_active():
            value = 'True'
        else:
            value = 'False'
        self.__setConfig('show_3d', value)

    def _on_show_ai_stats_clicked(self, widget, data = None):
        """Gtk+ callback"""
        window = self._gui.get_widget('ai_window')
        if widget.get_active():
            window.show()
        else:
            window.hide()

    def _on_history_combo_changed(self, widget, data = None):
        """Gtk+ callback"""
        model = widget.get_model()
        iter = widget.get_active_iter()
        if iter is None:
            return
        
        # Get the move number
        moveNumber = model.get_value(iter, 0)
        
        string = 'Show move number: ' + str(moveNumber)
        if moveNumber == len(model) - 1:
            string += ' (latest)'
            moveNumber = -1
        
        view = self.notebook.getView()
        if view is not None:
            view._setMoveNumber(moveNumber)
        
    def __selectMoveNumber(self, moveNumber):
        """FIXME
        """
        combo = self.__getWidget('history_combo')
        
        # Limit moves to the maximum value
        maxNumber = len(combo.get_model())

        # Allow negative indexing
        if moveNumber < 0:
            moveNumber = maxNumber + moveNumber
        if moveNumber < 0:
            moveNumber = 0
        if moveNumber >= maxNumber:
            moveNumber = maxNumber - 1
        
        combo.set_active(moveNumber)

    def __selectMoveNumberRelative(self, offset):
        """FIXME
        """
        combo = self.__getWidget('history_combo')
        selected = combo.get_active()
        maxNumber = len(combo.get_model())
        new = selected + offset
        if new < 0:
            new = 0
        elif new >= maxNumber:
            new = maxNumber - 1
        self.__selectMoveNumber(new)

    def _on_history_start_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumber(0)

    def _on_history_previous_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumberRelative(-1)

    def _on_history_next_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumberRelative(1)

    def _on_history_latest_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumber(-1)

    def _on_view_changed(self, widget, page, pageNum, data = None):
        """Gtk+ callback"""
        self._updateViewButtons()
    
    def _updateViewButtons(self):
        """
        """
        view = self.notebook.getView()
        enableWidgets = (view is not None) and view.isActive
        self.__getWidget('end_game_button').set_sensitive(enableWidgets)
        self.__getWidget('save_game_button').set_sensitive(enableWidgets)
        self.__getWidget('menu_save_item').set_sensitive(enableWidgets)
        self.__getWidget('menu_end_game_item').set_sensitive(enableWidgets)

        combo = self.__getWidget('history_combo')
        if view is None:
            combo.set_model(None)
        else:
            (model, selected) = view._getModel()
            combo.set_model(model)
            if selected < 0:
                selected = len(model) + selected
            combo.set_active(selected)
        self.__getWidget('navigation_box').set_sensitive(enableWidgets)

    def _on_new_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        dialogs.GtkNewGameDialog(self, self.__playerModel)

    def _on_join_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        # Create the dialog
        dialog = dialogs.GtkJoinGameDialog(self, self.__playerModel)
        self.__joinGameDialogs.append(dialog)
        # FIXME: Remove from this list when they dissapear
        
        # Add the detected games into the dialog
        for (game, name) in self.__networkGames.iteritems():
            dialog.addNetworkGame(name, game)

    def _on_open_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        dialogs.GtkLoadGameDialog(self)
        
    def _on_save_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        view = self.notebook.getView()
        if not self.__saveGameDialogs.has_key(view):
            self.__saveGameDialogs[view] = dialogs.GtkSaveGameDialog(self, view)
        
    def _on_end_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        view = self.notebook.getView()
        assert(view is not None)
        if view.feedback is not None:
            view.feedback.close()

    def _on_about_clicked(self, widget, data = None):
        """Gtk+ callback"""
        if self.__aboutDialog is None:
            self.__aboutDialog = gtk.glade.XML(UI_DATA_DIR + 'about.glade', domain = 'glchess')
            self.__aboutDialog.signal_autoconnect(self)
        
    def _on_glchess_about_dialog_close(self, widget, data = None):
        """Gtk+ callback"""
        self.__aboutDialog = None
        
    def _on_ai_window_delete_event(self, widget, data = None):
        """Gtk+ callback"""
        self._gui.get_widget('menu_view_ai').set_active(False)
        
        # Stop the event - the window will be closed by the menu event
        return True

    def _on_quit(self, widget, data = None):
        """Gtk+ callback"""
        self.onQuit()

if __name__ == '__main__':
    ui = GtkUI()
    ui.run()
